# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xdi', 'xdi._common']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.4.0,<22.0.0', 'typing-extensions>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'xdi',
    'version': '0.4.3',
    'description': 'Dependency injection library',
    'long_description': '# XDI\n\n\n[![PyPi version][pypi-image]][pypi-link]\n[![Supported Python versions][pyversions-image]][pyversions-link]\n[![Build status][ci-image]][ci-link]\n[![Coverage status][codecov-image]][codecov-link]\n\n\n`XDI` is a [dependency injection](https://en.wikipedia.org/wiki/Dependency_injection) library for Python.\n\n\n## Why Use XDI?\n\n- Fast: minus the cost of an additional stack frame, `xdi` resolves dependencies \nnearly as efficiently as resolving them by hand.\n- Async support: `xdi` will `await` for you.\n- Lots Providers to choose from.\n\n\n## Why use dependency injection?\n\nTake a look at the following code.\n\n```python\nimport os\n\nclass ApiClient:\n\n    def __init__(self):\n        self.api_url = os.getenv("API_URL")  # a dependency\n        self.api_key = os.getenv("API_KEY")  # a dependency\n\n    def get(self, path: str, *, token: str):\n        return { url: f\'{self.api_url}/{path}\', status: \'ok\', data: [] }\n\n\n\nclass Service:\n\n    def __init__(self):\n        self._api_client = ApiClient()  # a dependency\n        # do some init\n\n    def do_something(self):\n        res = self._api_client.get(\'abc\')\n        print("Service doing something")\n\n\ndef some_func(*args, **kwargs) -> None:\n    service = Service()  # a dependency\n    service.do_something()\n    print("serivce has done something")\n\n\nif __name__ == "__main__":\n    some_func()\n\n```\n\nThis code will run as expected. However:-\n\n1. **Testing it will be difficult**\n    \n    For example, to test `Service` we need a fake `ApiClient` as we don\'t what to \n    make real api calls or might not have the credentials to do so. \n    Since `Service` creates it\'s own `ApiClient` instance, it is impossible to \n    safely mock the `ApiClient` for tests. \n\n2. **Lacks flexibility and extensibility**\n\n    It\'s imppossible to create an additional `ApiClient` instance that uses a \n    different `API_URL` and/or `API_KEY`.\n\n\n\nSo what do we do? We should decouple our objects from their dependencies. \n\nThat is, objects should not create each other anymore. They should provide a way \nto inject the dependencies instead.\n\n\nHere\'s how.\n\n```python\n\nclass ApiClient:\n\n    def __init__(self, api_url: str, api_key: str): # we let the caller provide the dependencies\n        self.api_url, self.api_key = api_url, api_key\n        \n\nclass Service:\n\n    def __init__(self, api_client: ApiClient): # we let the caller provide the dependency\n        self._api_client = api_client \n\n\ndef some_func(*args, service: Service, **params): # we let the caller provide the dependency\n    ...\n\n\n```\nCongratulations, your code is now loosely coupled. \n\nBut remember, with freedom comes more responsibility.\n\nThe responsibility is left to the "caller" who has to know, assemble and provide the dependencies.\n\n```python\nsome_func(\n    *args, \n    service=Service(\n        api_client=ApiClient(\n            api_key=os.getenv("API_KEY"),\n            api_url=os.getenv("TIMEOUT"),\n        ),\n    ),\n    **kwargs, \n)\n\n```\nThis quickly becomes a problem when you what to use `some_func()` from multiple places.\nDuplicating the assembly code with make it harder to change in the future.\n\n\n\n\n### With XDI.\n\nSimple DI using `xdi`\'s low level API.\n\n```python\nfrom xdi import Injector, Scope, Container\n\ncontainer = Container()\n\n# register the ApiClient\ncontainer.singleton(\n    ApiClient, \n    api_url=os.getenv("API_URL"), # <-- provide value from env\n    api_key=os.getenv(\'API_KEY\')  # <-- provide value from env\n)  # <-- make it provide only one instance.\n\n\n# Register the Service. \ncontainer.factory(Service) \n# Since we did not specify a value for api_client. `ApiClient` will get injected \n# as it matches the type annotation `api_client: ApiClient`\n\n# Create the scope\nscope = Scope(container)\n\n# Create an injector\ninjector = Injector(scope)\n\n\n# Use the injector to run `some_func`\nresult = injector.make(some_func, \'xyz\', 23, **params) # <-- dependency `Service` is injected automatically\n\n```\n\n\n\n## Installation\n\nInstall from [PyPi](https://pypi.org/project/xdi/)\n\n```\npip install xdi\n```\n\n## Documentation\n\nFull documentation is available [here][docs-link].\n\n\n\n## Production\n\n__This package is still in active development and should not be used in production environment__\n\n\n\n\n[docs-link]: https://pyxdi.github.io/xdi/\n[pypi-image]: https://img.shields.io/pypi/v/xdi.svg?color=%233d85c6\n[pypi-link]: https://pypi.python.org/pypi/xdi\n[pyversions-image]: https://img.shields.io/pypi/pyversions/xdi.svg\n[pyversions-link]: https://pypi.python.org/pypi/xdi\n[ci-image]: https://github.com/pyxdi/xdi/actions/workflows/workflow.yaml/badge.svg?event=push&branch=master\n[ci-link]: https://github.com/pyxdi/xdi/actions?query=workflow%3ACI%2FCD+event%3Apush+branch%3Amaster\n[codecov-image]: https://codecov.io/gh/pyxdi/xdi/branch/master/graph/badge.svg\n[codecov-link]: https://codecov.io/gh/pyxdi/xdi\n\n\nSee this release on GitHub: [v0.4.3](https://github.com/pyxdi/xdi/releases/tag/0.4.3)\n',
    'author': 'David Kyalo',
    'author_email': 'davidmkyalo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pyxdi/xdi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
